
%lex

%s codes trans

%%

<INITIAL>[\s\S]+?(?="#(")		{ this.pushState('trans'); return 'TEXT'; }
<INITIAL>[\s\S]+				{ return 'TEXT'; }

<trans>"#"(?="(")				{ return 'TOPIC_SIGILS'; }
<trans>"("						{ this.popState(); this.pushState('codes'); return 'DOWN'; }

<codes>[\$@#]+(?=[\r\n\s)])		{ return 'TOPIC_SIGILS_EMPTY'; }
<codes>[\$@#]+					{ return 'TOPIC_SIGILS'; }
<codes>"("						{ this.pushState('codes'); return 'DOWN'; }

<codes>"\""{3}[\s\S]*?"\""{3,}	{ return 'QUOTES3'; }
<codes>"'"{3}[\s\S]*?"'"{3,}	{ return 'QUOTES3'; }

<codes>"\""(?:[^"\x5c]|\\[\s\S])*?"\""		{ return 'QUOTES1'; }
<codes>"'"(?:[^'\x5c]|\\[\s\S])*?"'"		{ return 'QUOTES1'; }

<codes>[^\r\n\s)]+				{ return 'QUOTES0'; }

<codes>")"						{ this.popState(); return 'UP'; }
<codes>[\r\n\s]+				/* skip */


/lex

%start top

%%

top
	: { $$ = []; return $$; }
	| items { $$ = $1; return $$; }
	;

items
	: item { $$ = [$1]; }
	| items item { $$ = $1.concat([$2]); }
	;

item : literal | string | node ;

literal
	: TEXT { $$ = MARKUP.literalText($1); }
	;

string
	: string_nosigil { $$ = $1; }
	| TOPIC_SIGILS string_nosigil { $$ = MARKUP.withSigils($1,$2); }
	;

string_nosigil
	: QUOTES3 { $$ = MARKUP.quote3($1.slice(3,-3), $1.slice(0,3)); }
	| QUOTES1 { $$ = MARKUP.quote1($1.slice(1,-1), $1.slice(0,1)); }
	| QUOTES0 { $$ = MARKUP.quote0($1); }
	;

node
	: node_nosigil
	| TOPIC_SIGILS node_nosigil { $$ = MARKUP.withSigils($1,$2); }
	| TOPIC_SIGILS_EMPTY { $$ = MARKUP.withSigils($1, MARKUP.quoteImpliedEmpty()); }
	;

node_nosigil
	: DOWN items UP { $$ = MARKUP.node($2) }
	;


%%

var MARKUP = {
	each : function(map, fn) {
		for(var k in map) {
			if(Object.prototype.hasOwnProperty.call(map, k)) {
				if(fn(map[k], k) === false) break;
			}
		}
	},
	tag : function(name, data, more) {
		return MARKUP.tagObject(name, data, more);
	},
	tagObject : function(name, data, more) {
		var rv = {};
		MARKUP.each(more, function(v,k) {
			rv[k] = v;
		});
		if(String(name) === name)
			rv[" NAME"] = name;
		if(!(data === null || data === void(0)))
			rv[" DATA"] = data;
		return rv;
	},
	withOneSigil : function(sigil, item) {
		return MARKUP.tag("sigils", item, { sigil : sigil });
	},
	withSigils : function(sigils, item) {
		return (sigils.length <= 1) ?
			MARKUP.withOneSigil(sigils, item) :
			MARKUP.withOneSigil(sigils.slice(0,1), MARKUP.withSigils(sigils.slice(1), item)); 
	},
	quote3 : function(rawText, quote) {
		return MARKUP.tag("string", rawText, { quote : quote });
	},
	quote1 : function(rawText, quote) {
		var text = rawText.replace(/\\(?:x([0-9A-Fa-f]{2})|u([0-9A-Fa-f]{4})|([\S\s]))/g, function(match, xc, uc, sc, offset, string) {
			if(String(sc) === sc) {
				return sc;
			}
			else {
				var h = parseInt( (String(xc) === xc) ? xc : uc, 16 );
				return String.fromCharCode(h);
			}
		});
		return MARKUP.tag("string", text, { quote : quote, rawText : rawText });
	},
	quote0 : function(rawText) {
		return MARKUP.tag("string", rawText);
	},
	quoteImpliedEmpty : function() {
		return MARKUP.tag("string");
	},
	literalText : function(text) {
		return MARKUP.tag("literal", text);
	},
	node : function(list) {
		return MARKUP.tag("node", list);
	}
};
